if {[info exists argv] && ([string first -testTXMLOO $argv] >= 0)} {
  ::tcl::tm::path add .
}

package require formatTextWin

oo::class create formatTextWinXML {
  super formatTextWin
################################################################
#   classmethod unknown {w args}--
#    Handles unknown event
# Arguments
#  w	First argument to unknown. Hopefully a window name
#  args	Everything else
#
# Results
#   If w is a window name, control is passed to class create method
#   and a new widget is created.
# 
  classmethod unknown {w args} {
    if {[string match .* $w]} {
      set newObj [[self] new $w {*}$args]
#      rename $w ${w}_WIN
      return $newObj
    }
    next $w {*}$args
  }

################################################################
#   method export {}--
#    Export the text widget contents as XML
#    Formatted per
# http://officeopenxml.com/WPtextFormatting.php
# Arguments
#   NONE	- Reads values from text widget
# 
# Results
#   Returns XML string.
# 
  method export {} {
    variable State
    set dumpData [next]

    lassign $dumpData tags tagDump textDump
    
    set skips {sel current insert}

    foreach {tt txt index} $textDump {
      append str $txt
    }
    
    set tag(sel) {}
    set untag(sel) {}
    
    set sawFont 0
    foreach {nm config} $tags {
      # puts "NM: $nm CFG: $config"
      foreach {k v} $config {
        # puts "  k: $k V: $v"
        switch $k {
	  -font {
	    lassign $v family size
	    if {$sawFont} {
	      append tag($nm) "</w:sz>"
	      append tag($nm) "</w:font>"
	    }
	    append tag($nm) "<w:sz w:val=\"$size\">"
	    append tag($nm) "<w:font w:name=\"$family\">"

	    if {[string first bold $v] >= 0} {
	      append tag($nm) "<w:b>"
	      append untag($nm) "</w:b>"
	    }
	    if {[string first italic $v] >= 0} {
	      append tag($nm) "<w:i>"
	      append untag($nm) "</w:i>"
	    }
	    set sawFont 1
	  }
	  -foreground {
	    append tag($nm) "<w:color w:val=\"$v\">"
	    # append tag($nm) "</w:color w:val=\"$v\">"
	    append untag($nm) "</w:color>"
	  }
	  -background {
	  }
	  -underline {
	    append tag($nm) "<w:u w:val=\"true\"></w:u>"
	    append untag($nm) "<w:u w:val=\"false\"></w:u>"
	  }
	  -overstrike {
	    append tag($nm) "<w:strike w:val=\"true\"></w:strike>"
	    append untag($nm) "<w:strike w:val=\"false\"></w:strike>"
	  }
	  default {
	    puts "UNKNOWN: $k"
	  }
	}
      }
      # puts "TAG $nm -> $tag($nm)"
    }

  set lineLst [split [$State(textWin) get 0.0 end] \n]

  foreach {index nm op} [lreverse $tagDump] {
    lassign [split $index .] ln ch
    incr ln -1
    incr ch -1
    set ll [lindex $lineLst $ln]
    switch $op {
      tagon {
        set ll [string range $ll 0 $ch]$tag($nm)[string range $ll [expr {$ch+1}] end]
      }
      tagoff  {
        set ll [string range $ll 0 $ch]$untag($nm)[string range $ll [expr {$ch+1}] end]
      }
    }
    # puts "B4 $ll -> $lineLst"
    set lineLst [lreplace $lineLst $ln $ln $ll]
    # puts "Post $ll -> $lineLst"
  }
  return $lineLst
  }

################################################################
#   method import {xml_data}--
#    Import XML data from export function into text widget
# Arguments
#   xml_data	XML string
# 
# Results
#   Converts xml into text widget format and 
#   invokes parent import function to modify widget.
# 
  method import {xml_data} {
    variable State
    package require htmlparse
    
    set ln 0
    set State(txt) ""
    set State(tagDef) ""
    set State(tagInfo) ""
    set State(tagNum) 1
    set State(Config) ""
    
    foreach option {fontFam fontSiz bold italic underline overstrike color} {
      set State(tagconfig,$option) ""
      set State(stack,$option) ""
    }
    foreach nm [array names State tagconfig,*\[0-9\]] {
      unset State($nm)
    }


    foreach line $xml_data {
      incr ln
      # puts "LINE: $line"
      set rtn [htmlparse::parse -cmd [list [self] parseMethod $ln] <html>$line</html>]
      append State(txt) "\n"
    }

    my defineTag

    if {$State(tagNum) > 1} {
        lappend State(tagOnOff) tagoff tag_[expr {$State(tagNum) -1}] [list [list 1.0 + [string length $State(txt)] chars]]
      }
      lappend State(tagOnOff) tagon tag_$State(tagNum) [list [list 1.0 + [string length $State(txt)] chars]]
  
     set l1 $State(Config)
     set l3 [list text $State(txt) end]
     set l2 $State(tagOnOff)
     
     if {0} {
       puts "OUT: "
       foreach {k v} $l1 {
         puts "$k $v"
       }
       foreach {to tn po} $l2 {
         puts "$to $tn $po"
       }
       puts "$l3"
     }
     next [list $l1 $l2 $l3]
  }

################################################################
#   method parseMethod {ln a b c d}--
#    Parse a XML tag
# Arguments
#   ln		Line number 
#   a		tag name (w:b, w:i, etc
#   b		On/Off <w:b> -> "" </w:b> -> "/"
#   c		Other part of tag: </w:color val="red"> -> val="red"
#   d		Text following tag. This marks the end of a tag sequence.
# 
# Results
#   
# 
  method parseMethod {ln a b c d} {
    variable State
    # puts "LN: $ln A: $a B: $b C: $c D: $d"
    # ln - line num
    # a: cmd 
    # b: / or no / to mark close
    # c: argument
    # d: text between HTML tags

    set tagDef ""
    set tagInfo ""
    set txt [lindex [split $State(txt) \n] end]

    switch $a {
      w:sz {
        set val [my recordChange fontSiz $ln $b $c]
      }
      w:font {
        set val [my recordChange fontFam $ln $b $c]
      }
      w:color {
        set val [my recordChange color $ln $b $c]
      }
      w:b {
        set val [my recordToggle bold $ln $b bold]
      }
      w:i {
        set val [my recordToggle italic $ln $b italic]
      }
      w:strike {
        set val [my recordChange overstrike $ln $b $c]
      }
      w:u {
        set val [my recordChange underline $ln $b $c]
      }
      hmstart {
      }
      html {
      }
      default {
        error "Unknown tag: $a ($b $c $d)"
      }
    }
    if {$d ne ""} {
      my defineTag
      if {$State(tagNum) > 1} {
        lappend State(tagOnOff) tagoff tag_[expr {$State(tagNum) -1}] [list [list 1.0 + [string length $State(txt)] chars]]
      }
      lappend State(tagOnOff) tagon tag_$State(tagNum) [list [list 1.0 + [string length $State(txt)] chars]]
      incr State(tagNum)
      append State(txt) $d
    }
  }
  
################################################################
#   method defineTag {}--
#    Examine current stack and update tag configuration
# Arguments
#   NONE - Reads from State(stack,*)
# 
# Results
#   State(Config) is extended with tag name and configuration
# 
  method defineTag {} {
    variable State
    set font [my getFont]
    set tagDef [list -font $font]
    set foreground [my peek color]
    if {$foreground ne ""} {
      lappend tagDef -foreground $foreground
    }
    if {[my peek underline] ne ""} {
      lappend tagDef -underline [my peek underline]
    }
    if {[my peek overstrike] ne ""} {
      lappend tagDef -overstrike [my peek overstrike]
    }
    lappend State(Config) tag_$State(tagNum) $tagDef
  }

################################################################
#   method getFont {}--
#    Examine stacks and determine font definition.
# Arguments
#   NONE - Reads from State(stack,*)
# 
# Results
#   No side effects
# 
  method getFont {} {
    variable State

    if {[catch {my peek fontFam} fam]} {set fam ""}
    if {[catch {my peek fontSiz} siz]} {set siz ""}
    if {[catch {my peek bold}    bold]} {set bold ""}
    if {[catch {my peek italic}  italic]} {set italic ""}
    set def "$fam $siz $italic $bold"
    return $def
  }

################################################################
#   method getval {str}--
#    split val="VALUE" and return VALUE without quotes
# Arguments
#   str		$c from Parse - 'val="CONTENTS"'
# 
# Results
#   Returns value, not quotes
# 
  method getval {str} {
    lassign [split $str =] op val
    return [string trim $val \"]
  }

################################################################
#   method recordChange {id ln b c}--
#    Update stacks depending on b and c
#    '<w:color val="red">' uses recordChange
#    '<w:b></w:b> uses recordToggle
# Arguments
#   id	stack to update (bold, fontSiz, etc)
#   ln  Line number - to track index for tag
#   b   From parse "/" or ""
#   c   From parse - 'val="value"'
# 
# Results
#   Pushes or pops stack $id, pushes tagOn/tagOff stack
# 
  method recordChange {id ln b c} {
    variable State
    set val ""

    set txt [lindex [split $State(txt) \n] end]

    # puts "RC: ID: $id LN: $ln B: $b C: $c "
    # catch {
    # puts "RC: b4: $State(stack,$id)"
    # }
   if {$b ne "/"} {
      set val [my getval $c]
    # puts "VAL2: $val"
      my push $id $val
      lappend State(tagconfig,$id) $val
      my pushnew tagOn $ln.[string length $txt]
    } else {
      my pop $id
      set val [my peek $id]
      my pushnew tagOff $ln.[string length $txt]
    }
    # puts "RC: END: $State(stack,$id)"
    return $val
  }

################################################################
#   method recordToggle {id ln b c}--
#    Update stacks for a toggle type XML tag.
#    '<w:color val="red">' uses recordChange
#    '<w:b></w:b> uses recordToggle
# Arguments
#   id	stack to update (bold, fontSiz, etc)
#   ln  Line number - to track index for tag
#   b   From parse "/" or ""
#   c   From parse - 'val="value"'
# 
# Results
#   Pushes or pops stack $id, pushes tagOn/tagOff stack
# 
  method recordToggle {id ln b c} {
    variable State
    set val ""
    set txt [lindex [split $State(txt) \n] end]

    # puts "RT: ID: $id LN: $ln B: $b C: $c "
    # catch {
    # puts "RC: b4: $State(stack,$id)"
    # }
    if {$b eq "/"} {
      my pop $id
      set val [my peek $id]
    # puts "VAL: $val"
      my pushnew tagOff $ln.[string length $txt]
      if {$val ne ""} {
        my pushnew tagOn $ln.[string length $txt]
        lappend State(tagconfig,$id) "$val"
      }
    } else {
        set val $c
    # puts "VAL2: $val"
      my push $id $val
      lappend State(tagconfig,$id) $val
      my pushnew tagOn $ln.[string length $txt]
    }
    # puts "RC: END: $State(stack,$id)"
    return $val
  }

################################################################
#   method pushnew {stack val}--
#    Only push a value if it's different from top of stack
#    Mostly for tagOn/tagOff. There's no knowing how many XML tags
#    will be recorded, but it will result in a single textWidget tag
#    configure.
# Arguments
#   stack	Name of stack to modify
#   val		Value to push
# 
# Results
#   Stack may be modified
# 
  method pushnew {stack val} {
    variable State
    if {(![info exists State(stack,$stack)]) || 
        ([lindex $State(stack,$stack) end] ne $val)} {
      lappend State(stack,$stack) $val
    }
  }

################################################################
#   method push {stack val}--
#    Standard stack push operation
# Arguments
#   stack	Name of stack
#   val		Value to push
# 
# Results
#   Adds value to end of stack
# 
  method push {stack val} {
    variable State
    lappend State(stack,$stack) $val
  }

################################################################
#   method pop {stack }--
#    Standard stack pop operation
# Arguments
#   stack	Name of stack
# 
# Results
#   Stack is shortened. Value is returned.
# 
  method pop {stack } {
    variable State
    set rtn [lindex $State(stack,$stack) end]
    set State(stack,$stack) [lrange $State(stack,$stack) 0 end-1]
    return $rtn
  }

################################################################
#   method peek {stack}--
#    Examine top of stack without change
# Arguments
#   stack	Name of stack
# 
# Results
#   No changes
# 
  method peek {stack} {
    variable State
    set rtn [lindex $State(stack,$stack) end]
    return $rtn
  }
}



if {[info exists argv] && ([string first -testTXMLOO $argv] >= 0)} {
  ::tcl::tm::path add .
  package require formatTextWin

  
  if {[string first -tktest $argv] >= 0} {
    source ../tktest/socksend/socksend.tcl
    sockappsetup tktest.tcl 3010  127.0.0.1
  }

  destroy {*}[winfo children .]
  toplevel .bb
  set obj1 [formatTextWinXML .txt -controlFrame .bb -fontFamily courier \
      -fontSize 14]
  puts "OBJ: $obj1"
  grid [$obj1 configure -textWin] -columnspan 4
  catch {
    grid [$obj1 configure -controlFrame]
  }
  
  set obj2 [formatTextWinXML .txt2  -fontFamily courier \
      -fontSize 14 -controlStyle row]
  puts "OBJ2: $obj2"
  grid [$obj2 configure -textWin ] -columnspan 4
  grid [$obj2 configure -controlFrame]  -columnspan 4
  
  button .bb1 -text "Export from top page" -command {set data [$obj1 export]}
  button .bb2 -text "Import into bottom page" -command {$obj2 import $data}
  button .bb3 -text "save ExportData" -command {
    set of [open save.dat w]
    puts $of $data
    close $of
  }
  button .bb4 -text "load ExportData" -command {
    set if [open save.dat r]
    set data [read $if]
    close $if
  }
  grid .bb1 .bb2 .bb3 .bb4

  toplevel .cc
  set obj3 [formatTextWin .txt3  -fontFamily courier -controlFrame .cc\
      -fontSize 14 -controlStyle row]
  puts "obj3: $obj3"
  grid [$obj3 configure -textWin ] -columnspan 4

  
}

